/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/*!
@file imageIO.cpp
@ingroup ExamplePrograms
@brief Simple example of image input/output from a video file, to a canvas window

This program will display in an image canvas the contents of a video file or the input from a video resource.

@section Usage Usage of the program.
Once installed the QVision, compile using the <i>qmake</i> and <i>make</i> tools:
@code # qmake; make @endcode
Execute the application with the following line:
@code ./imageIO --URL=<source video identifier string> @endcode
For example:
@code ./imageIO --URL=http://perception.inf.um.es/public_data/videos/misc/minuto.avi @endcode
See section @ref QVMPlayerURLFormat for further info about the format and types of source video identifier strings.
@author PARP Research Group. University of Murcia, Spain.
*/

#include <QVApplication>
#include <QVMPlayerCamera>
#include <QVDefaultGUI>
#include <QVImageCanvas>
#include <QVVideoRecorder>

/////////////////////////////////////////////////////////////////////////////////////////////////////////
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#define PNG_DEBUG 3
#include <png.h>

void abort_(const char * s, ...)
{
	va_list args;
	va_start(args, s);
	vfprintf(stderr, s, args);
	fprintf(stderr, "\n");
	va_end(args);
	abort();
}

int x, y;

int width, height;
png_byte color_type;
png_byte bit_depth;

png_structp png_ptr;
png_infop info_ptr;
int number_of_passes;
png_bytep * row_pointers;

void read_png_file(const char* file_name, QVImage<uChar, 3> &image)
	{
	png_byte header[8];	// 8 is the maximum size that can be checked
	
	/* open file and test for it being a png */
	FILE *fp = fopen(file_name, "rb");
	if (!fp)
		abort_("[read_png_file] File %s could not be opened for reading", file_name);
	fread(header, 1, 8, fp);
	if (png_sig_cmp(header, 0, 8))
		abort_("[read_png_file] File %s is not recognized as a PNG file", file_name);
		
	/* initialize stuff */
	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	
	if (!png_ptr)
		abort_("[read_png_file] png_create_read_struct failed");
	
	info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr)
		abort_("[read_png_file] png_create_info_struct failed");
	
	if (setjmp(png_jmpbuf(png_ptr)))
		abort_("[read_png_file] Error during init_io");
	
	png_init_io(png_ptr, fp);
	png_set_sig_bytes(png_ptr, 8);
	
	png_read_info(png_ptr, info_ptr);
	
	width = info_ptr->width;
	height = info_ptr->height;

	image = QVImage<uChar, 3>(width, height);

	color_type = info_ptr->color_type;
	bit_depth = info_ptr->bit_depth;
	
	number_of_passes = png_set_interlace_handling(png_ptr);
	png_read_update_info(png_ptr, info_ptr);
	
	
	/* read file */
	if (setjmp(png_jmpbuf(png_ptr)))
		abort_("[read_png_file] Error during read_image");
	
	row_pointers = (png_bytep*) malloc(sizeof(png_bytep) * height);
	for (y=0; y<height; y++)
		//row_pointers[y] = (png_byte*) malloc(info_ptr->rowbytes);
		row_pointers[y] = image.getWriteData() + y*image.getStep();

	png_read_image(png_ptr, row_pointers);
	
	fclose(fp);

	const int Channels = 4;
	for (int i= 0; i < height; i++)
		for (int j = 0; j < width; j++)
			for (int c = 0; c < 3; c++)
				image(j,i,c) = row_pointers[i][Channels*j+c];
	}


void write_png_file(const char* file_name)
{
	/* create file */
	FILE *fp = fopen(file_name, "wb");
	if (!fp)
		abort_("[write_png_file] File %s could not be opened for writing", file_name);


	/* initialize stuff */
	png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	
	if (!png_ptr)
		abort_("[write_png_file] png_create_write_struct failed");

	info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr)
		abort_("[write_png_file] png_create_info_struct failed");

	if (setjmp(png_jmpbuf(png_ptr)))
		abort_("[write_png_file] Error during init_io");

	png_init_io(png_ptr, fp);


	/* write header */
	if (setjmp(png_jmpbuf(png_ptr)))
		abort_("[write_png_file] Error during writing header");

	png_set_IHDR(png_ptr, info_ptr, width, height,
		     bit_depth, color_type, PNG_INTERLACE_NONE,
		     PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

	png_write_info(png_ptr, info_ptr);


	/* write bytes */
	if (setjmp(png_jmpbuf(png_ptr)))
		abort_("[write_png_file] Error during writing bytes");

	png_write_image(png_ptr, row_pointers);


	/* end write */
	if (setjmp(png_jmpbuf(png_ptr)))
		abort_("[write_png_file] Error during end of write");

	png_write_end(png_ptr, NULL);

        /* cleanup heap allocation */
	for (y=0; y<height; y++)
		free(row_pointers[y]);
	free(row_pointers);

        fclose(fp);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef DOXYGEN_IGNORE_THIS
class PlayerWorker: public QVWorker
	{
	public:
		PlayerWorker(QString name): QVWorker(name)
			{
			addProperty< QVImage<uChar,3> >("Input image", inputFlag|outputFlag);
			addProperty< QString >("Image name", inputFlag, "flower.ppm");
			QVImage<uChar,3> image;

			readQVImageFromFile(getPropertyValue< QString > ("Image name"), image);
			//read_png_file(qPrintable(getPropertyValue< QString > ("Image name")), image);
			addProperty< QVImage<uChar,3> >("Flower", outputFlag, image);

			writeQVImageToFile("flower2.ppm", image);
			}

		void iterate()	{ }
	};

int main(int argc, char *argv[])
	{
	QVApplication app(argc, argv, "Example program for QVision library. Displays the contents of a video source.");

	QVMPlayerCamera camera("Video");
	PlayerWorker player("Video player");
	camera.link(&player,"Input image");
	//camera.link(&player,"Input image Y", "Input image U", "Input image V");

	QVVideoRecorder recorder("Video recorder");
	//recorder.linkProperty(player, "Input image Y", "Input image U", "Input image V");
	recorder.linkProperty(player, "Input image");

	QVDefaultGUI interface;

	QVImageCanvas imageCanvas("Output image");
	imageCanvas.linkProperty(player, "Input image");
	/*QVImageCanvas imageCanvasY("Output image Y");
	imageCanvasY.linkProperty(player, "Input image Y");
	QVImageCanvas imageCanvasU("Output image U");
	imageCanvasU.linkProperty(player, "Input image U");
	QVImageCanvas imageCanvasV("Output image V");
	imageCanvasV.linkProperty(player, "Input image V");*/

	QVImageCanvas imageCanvasTest("flower.ppm");
	imageCanvasTest.linkProperty(player, "Flower");

	return app.exec();
	}

#endif

